#!/usr/bin/ruby
#
# Copyright (C) 2003, 2004, 2005, 2014, 2015 Filip Pizlo. All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY FILIP PIZLO ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL FILIP PIZLO OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
#

puts "gpc_gen.rb -- generates a bunch of instruction-related stuff for"
puts "              the GPC by reading the single instruction spec."

$input = File.open(ARGV.shift)
$outputDir = ARGV.shift

# Some nice little utility methods

$lines_parsed=0
def parseline
  while line = $input.gets
    $lines_parsed += 1
    line.chomp!
    unless line =~ /^#/ or
        line =~ /^\s*$/
      return line
    end
  end
  nil
end

def isnumber(str)
  str =~ /^-?(0x[A-Fa-f0-9]+)|([0-9]+)$/
end

def tofstr(str)
  str.split("\n").collect{|x| x.inspect[1..-2].gsub(/%/,'%%')}.join("\n")
end

def parseparam(str)
  i=0
  balance=0
  param=''
  loop {
    if str[i]==?(
      balance+=1
    elsif str[i]==?)
      balance-=1
      break if balance<0
    end
    param << str[i]
    i+=1
  }
  [param, str[i+1..-1]]
end

def generate(file)
  puts ">> Generating #{file}..."
  Dir.chdir($outputDir) {
    File.open(file,'w') {
      | io |
      io.puts '/* generated by gpc_gen.rb */'
      yield io
    }
  }
end

# What we believe an instruction looks like

Instruction=Struct.new(:name,:stack,:stream,:atline,:code)

# Phase 1: parse in the instruction specification

set=[]

loop {
  line=parseline
  break unless line
  raise "Bad format: "+line unless line =~ /^\[([A-Z0-9_]+)\]$/
  instName=$1
  
  line=parseline
  raise "Bad format: "+line unless line =~ /^stack:\s*/
  stack=$'
  
  line=parseline
  raise "Bad format: "+line unless line =~ /^stream:\s*/
  stream=$'
  
  # now comes the actual code
  atline=$lines_parsed+1
  code=''
  loop {
    line=$input.gets
    $lines_parsed+=1
    if line=~/^[\t ]/
      code+="    "+line
    elsif line=~/^#/
      code+=line
    else
      break
    end
  }
  
  set << Instruction.new(instName,stack,stream,atline,code)
}

puts ">> Read #{set.size} instructions."

# Phase 2: generate a whole bunch of stuff!!

generate('gpc_instruction_static_size.gen') {
  | io |
  io.puts 'uint32_t gpc_instruction_static_size(gpc_cell_t inst) {'
  io.puts '    switch (inst) {'
  set.each {
    | inst |
    io.print "        case GPC_#{inst.name}: return "
    if isnumber(inst.stream)
      io.print "1+#{inst.stream}"
    else
      io.print 'UINT32_MAX'
    end
    io.puts ';'
  }
  io.puts '        default: break;'
  io.puts '    }'
  io.puts '    tsf_abort("Invalid instruction");'
  io.puts '}'
} 

generate('gpc_instruction_to_string.gen') {
  | io |
  io.puts 'const char *gpc_instruction_to_string(gpc_cell_t inst) {'
  io.puts '    switch (inst) {'
  set.each {
    | inst |
    io.puts "        case GPC_#{inst.name}: return "+
            "#{inst.name[2..-1].inspect};"
  }
  io.puts '        default: break;'
  io.puts '    }'
  io.puts '    tsf_abort("Invalid instruction");'
  io.puts '}'
}

generate('gpc_instruction_stack_effects.gen') {
  | io |
  io.puts 'int32_t gpc_instruction_stack_effects(gpc_cell_t inst) {'
  io.puts '    switch (inst) {'
  set.each {
    | inst |
    io.puts "        case GPC_#{inst.name}: return #{inst.stack};"
  }
  io.puts '        default: break;'
  io.puts '    }'
  io.puts '    tsf_abort("Invalid instruction");'
  io.puts '}'
}

generate('gpc_instruction_size.gen') {
  | io |
  io.puts '#define INT_STREAM(arg) (inst[1+(arg)])'
  io.puts 'uint32_t gpc_instruction_size(gpc_cell_t *inst) {'
  io.puts '    switch (*inst) {'
  set.each {
    | inst |
    next if isnumber(inst.stream)
    io.puts "        case GPC_#{inst.name}: return 1+#{inst.stream};"
  }
  io.puts '        default: break;'
  io.puts '    }'
  io.puts '    return gpc_instruction_static_size(*inst);'
  io.puts '}'
  io.puts '#undef INT_STREAM'
}

generate('gpc_instruction_dispatch.gen') {
  | io |
  first=true
  set.each {
    | inst |
    if first
      first=false
    else
      io.puts ","
    end
    io.print "&&#{inst.name}"
  }
  io.puts
}

generate('gpc_interpreter.gen') {
  | io |
  $atline=3
  def io.myputs(str)
    puts(str)
    $atline+=1
  end
  io.myputs '#define INT_STREAM_GOTO(arg) INT_GOTO(INT_STREAM(arg))'
  set.each {
    | inst |
    
    special_int_stream_goto=false
    
    if inst.code =~ /INT_STREAM_GOTO/ and
        inst.stack.to_i<0
      io.myputs '#undef INT_STREAM_GOTO'
      io.myputs '#define INT_STREAM_GOTO(arg) do {\\'
      (-inst.stack.to_i).times {
        io.myputs '    INT_POP();\\'
      }
      io.myputs '    INT_GOTO(INT_STREAM(arg));\\'
      io.myputs '} while (0)'
      special_int_stream_goto=true
    end
    
    io.myputs "INT_BEGIN(#{inst.name})"
    
    if inst.stack.to_i>0
      inst.stack.to_i.times {
        io.myputs '    INT_PUSH();'
      }
    end
    
    io.myputs "    {"
    io.myputs "#line #{inst.atline} \"gpc_spec.in\""
    inst.code.each_line {
      | line |
      io.myputs line
    }
    io.myputs "#line #{$atline+1} \"gpc_interpreter.gen\""
    io.myputs "    }"
    
    if inst.stack.to_i<0
      (-inst.stack.to_i).times {
        io.myputs '    INT_POP();'
      }
    end
    
    io.myputs "    INT_ADVANCE(#{inst.stream});"
    io.myputs 'INT_END()'
    
    if special_int_stream_goto
      io.myputs '#undef INT_STREAM_GOTO'
      io.myputs '#define INT_STREAM_GOTO(arg) '+
                'INT_GOTO(INT_STREAM(arg))'
    end
  }
  io.myputs '#undef INT_STREAM_GOTO'
}

def printstuff(line)
  format=''
  params=[]
  
  remaining=line
  while remaining=~/INT_/
    format+=tofstr($`)
    remaining=$'
    
    raise "Parse error" unless remaining=~/\(/
    
    cmd=$`
    param,remaining=parseparam($')
    
    if cmd=='STACK'
      format+='stack_" fptr "'
      params+=["height-#{param}-1"]
    elsif cmd=='STACK_SET'
      format+='stack_" fptr "='
      params+=["height-#{param.split(',')[0]}-1"]
      f,p=printstuff(param.split(',')[1])
      format+=f
      params+=p
    elsif cmd=='STACK_PTR'
      format+='((uint8_t*)stack_" fptr ")'
      params+=["height-#{param}-1"]
    elsif cmd=='STACK_PTR_SET'
      format+='stack_" fptr "=(gpc_cell_t)('
      params+=["height-#{param.split(',')[0]}-1"]
      f,p=printstuff(param.split(',')[1])
      format+=f
      params+=p
      format+=')'
    elsif cmd=='STREAM'
      if isnumber(param)
        format+='TSF_UINTPTR_C(" fuptr ")'
        params+=["cur[1+#{param}]"]
      else
        format+='values[" fuptr "+'
        params+=['cur-gpc_intable_get_stream(intable)+1']
        f,p=printstuff(param)
        format+=f
        params+=p
        format+=']'
      end
    elsif cmd=='STREAM_GOTO'
      if isnumber(param)
        format+='goto cell" fuptr "'
        params+=["cur[1+#{param}]"]
      else
        format+='LOOKUP_GOTO(" fuptr "+'
        params+=['cur-gpc_intable_get_stream(intable)+1']
        f,p=printstuff(param)
        format+=f
        params+=p
        format+=')'
      end
    elsif cmd=='FAILURE'
      format+='goto failure'
    elsif cmd=='CLEANUP'
      format+='(0)'
    else
      raise "#{cmd} is not a valid INT command"
    end
  end
  [format+tofstr(remaining),params]
end

