/*
 * Copyright (C) 2011 Filip Pizlo. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY FILIP PIZLO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL FILIP PIZLO OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

declare tsf_fsdb_buffer: array of uint8

declare tsf_fsdb_state: choice {
    /** the connection is ready for put/get requests. */
    clear: void
    
    /** the connection is in 'put' mode, and is expecting either
     * 'out_write', 'out_commit', or 'reset' commands. */
    put: void
    
    /** the connection is in 'get' mode, and is sending data
     * until EOF.  the only valid command is 'reset'. */
    get: void
}

/** command sent from client to server.  each command is acknowledged
 * by a response, which may be just 'ok'. */
declare tsf_fsdb_command: choice {
    /** no-op command, which illicits an ok response. */
    ping: void
    
    /** start putting a value at the key identified by the given
     * keysum, which must be an array of 5 uint32's.  this places
     * the connection into the 'put' state, and the only valid
     * commands are out_write, out_commit, or reset.
     * acknowledged with an 'ok' response or an error. */
    put: struct {
        key: any
    }
    
    /** write a sequence of bytes into the file opened by 'put'.
     * note that there need not be a correspondence between the
     * buffers from tsf_fsdb_out_write's perspective, and the
     * units of data sent here.  acknowledged with an 'ok' or
     * an error. */
    out_write: struct {
        data: tsf_fsdb_buffer
    }
    
    /** commit the 'put'.  regardless of success or failure, the
     * connection switches to 'clear' state.  acknowledged with
     * an 'ok' or an error. */
    out_commit: void
    
    /** start getting values at the key identified by the given
     * keysum, which must be an array of 5 uint32's.  this
     * places the connection into the 'get' state, and the only
     * valid command becomes 'reset'.  acknolwedged with
     * either 'ok', 'not_found', or an error.  if acknowledged
     * with 'ok', the server will start sending data.  the data
     * stream will end with either 'sys_error' or 'in_eof', at
     * which point the connection will be back in 'clear' state. */
    get: struct {
        key: any
    }
    
    /** remove an entry from the database. acknowledged with either
     * an 'ok', 'not_found', or an error. */
    rm: struct {
        key: any
    }
    
    /** reset the connection back to 'clear' state.  this can
     * be used to either abort a 'put' or to close a 'get'
     * prematurely.  if issued in 'put' or 'clear' states, it
     * is acknowledged immediately with an 'ok'.  if issued
     * in 'get' state, there may be some number of 'in_data'
     * or 'in_eof' packets queued up, as well as possibly a
     * 'sys_error', so be sure to either ignore or report those
     * and wait for the 'ok' before proceeding further.  note
     * that a somewhat safer method of aborting a put or get
     * is to simply disconnect from the server.  currently that
     * is faster for aborting a get, but slower for aborting
     * a put. */
    reset: void
}

/** response sent from server to client */
declare tsf_fsdb_response: struct {
    /** describes the state of the connection as a result of
     * the execution of the most recent command. */
    state: tsf_fsdb_state

    payload: choice {
        /** generic acknowledgement that the previous command
         * succeeded. */
        ok: void
        
        /** response sent to a new connection. */
        welcome: struct {
            /** the name of the server software. */
            name: string
            
            /** the version of the server software; this is usually
             * the TSF version. */
            version: struct {
                major: int32
                minor: int32
                patch: int32
            }
        }
    
        /** an error indicating an unexpected system fault. */
        sys_error: struct {
            msg: string
        }
        
        /** an error arising out of incorrect use of the protocol. */
        proto_error: struct {
            /** a server-generated message describing what went wrong */
            msg: string
        }
        
        /** indicates that the given key was not found in the
         * database. */
        not_found: void
    
        /** a data packet in response to a 'get' request.  these buffers
         * may not correspond to the buffers returned from
         * tsf_fsdb_in_read(), but may be at any granularity. */
        in_data: struct {
            data: tsf_fsdb_buffer
        }
        
        /** end-of-file on a 'get' request; indicates that all available
         * data has been sent. */
        in_eof: void
    }
}

