/*
 * Copyright (C) 2003, 2004, 2005, 2014 Filip Pizlo. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY FILIP PIZLO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL FILIP PIZLO OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "tsf.h"
#include "tsf_internal.h"
#include "tsf_format.h"
#include "write_cases.h"
#include "gpc.h"

#if defined(TSF_HAVE_MMAP) &&\
    defined(TSF_HAVE_MUNMAP) &&\
    defined(TSF_HAVE_MPROTECT) &&\
    defined(TSF_HAVE_GETPAGESIZE) &&\
    (defined(TSF_HAVE_MAP_ANON) ||\
     defined(TSF_HAVE_MAP_ANONYMOUS))
    
#include "tsf_test.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <stdarg.h>

#ifdef TSF_HAVE_MAP_ANON
#define TSF_MAP_ANON MAP_ANON
#else
#define TSF_MAP_ANON MAP_ANONYMOUS
#endif

static tsf_bool_t use_generation;

static void doit(tsf_reflect_t *data) {
    /* plan:
     * 1) generate a buffer
     * 2) repeat the following starting with size and going to 0:
     *   1) mmap some stuff and mprotect it.  use MAP_ANON
     *   2) copy the buffer into the mmap'd memory
     *   3) invoke either reflective or generation parser, depending on
     *      what we are configured to do
     *   4) munmap */
    
    tsf_buffer_t *buffer;
    int32_t size;
    size_t page_size = getpagesize();
    tsf_parser_t *parser = NULL;  /* make GCC happy */
    
    TAT(data);
    
    if (use_generation) {
        tsf_type_t *type = tsf_type_dup(tsf_reflect_get_type(data));
        
        TEST_ASSERT_TSFERROR(tsf_native_type_produce_some_mapping(&type,
								  tsf_true));
        
        TEST_ASSERT_TSFERROR(parser=tsf_parser_create(type));
        
        tsf_type_destroy(type);
    }
    
    TEST_ASSERT_TSFERROR(buffer =
        tsf_reflect_make_buffer(data));
    
    for (size = tsf_buffer_get_size(buffer);
         size >= 0;
         --size) {
        void *addr;
        void *fail_point;
        tsf_buffer_t crazy_buffer;
        
        TEST_ASSERT_ERRNO((addr=mmap(NULL,
                                     size + page_size * 2,
                                     PROT_READ | PROT_WRITE,
                                     TSF_MAP_ANON | MAP_PRIVATE,
                                     -1,
                                     0)) != (void*)-1);
        
        fail_point=(void*)((((uintptr_t)addr) + size + page_size - 1)
                           & ~(page_size - 1));
        
        TEST_ASSERT_ERRNO(mprotect(fail_point,
                                   page_size,
                                   0) >= 0);
        
        crazy_buffer.type = buffer->type;
        crazy_buffer.types = buffer->types;
        crazy_buffer.keep = tsf_false;
        crazy_buffer.data = ((uint8_t*)fail_point) - size;
        crazy_buffer.size = size;
        
        memcpy(crazy_buffer.data,
               buffer->data,
               size);
        
        if (use_generation) {
            void *data = tsf_parser_parse(parser, &crazy_buffer);
            if ((uint32_t)size == buffer->size) {
                TEST_ASSERT_TSFERROR(data);
                tsf_region_free(data);
            } else {
                TET(data, TSF_E_PARSE_ERROR);
            }
        } else {
            tsf_reflect_t *data = tsf_reflect_from_buffer(&crazy_buffer);
            if ((uint32_t)size == buffer->size) {
                TEST_ASSERT_TSFERROR(data);
                tsf_reflect_destroy(data);
            } else {
                TET(data, TSF_E_PARSE_ERROR);
            }
        }
        
        munmap(addr, size + page_size * 2);
    }
    
    if (use_generation) {
        tsf_parser_destroy(parser);
    }
    
    tsf_buffer_destroy(buffer);
    tsf_reflect_destroy(data);
}

int main(int c, char **v) {
    TEST_INIT(c, v);
    
    TEST_PUSH("reflect");
    use_generation = tsf_false;
    do_tests(tsf_false, doit);
    TEST_POP();
    
    TEST_PUSH("generation");
    use_generation = tsf_true;
    do_tests(tsf_false, doit);
    TEST_POP();
    
    TEST_DONE();
    
    return 0;
}

#else

int main(int c,char **v) {
    printf("%s: cannot do test because memory mapping and "
           "protecting doesn't work.\n", v[0]);
    return 0;
}

#endif

