/*
 * Copyright (C) 2004, 2005, 2015 Filip Pizlo. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY FILIP PIZLO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL FILIP PIZLO OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

typedef void create_data_t;

static tsf_buffer_t *create_make_buffer(tsf_type_t *type,
                                        void *data) {
    tsf_genrtr_t *gen;
    tsf_buffer_t *buf;
    TAT(type=tsf_type_dup(type));
    TAT(tsf_native_type_produce_some_mapping(&type,tsf_true));
    TAT(gen=tsf_generator_create(type));
    TAT(buf=tsf_generator_generate(gen,data));
    tsf_generator_destroy(gen);
    return buf;
}

static void create_destroy_data(void *data) {
    /* do nothing! :-) */

    /* NOTE: we could tsf_region_free(), in here, except that
       this would prevent easy embedding, since we would have to
       copy stuff from one region into another. */

    /* HMM: the above makes me think that having a region merge
       feature would be really freakin' cool.  And it shouldn't
       be too difficult to make. */
}

#define DEFINE_PRIMITIVE_CREATE_FUNCS(type_lc_name, type_lc, type_uc)   \
    static void *create_##type_lc_name(type_lc value) {                 \
        type_lc *result;                                                \
        TAE(result=malloc(sizeof(type_lc)));                            \
        *result=value;                                                  \
        return result;                                                  \
    }                                                                   \
    typedef int define_primitive_create_funcs_##type_lc_name

static void *create_bit(tsf_bool_t value) {
    tsf_bool_t *result;
    TAE(result=malloc(sizeof(tsf_bool_t)));
    *result=value;
    return result;
}

static void *create_void(void) {
    return NULL;
}

static void *create_string(uint32_t size) {
    char **string;
    uint32_t i;
    TAE(string=malloc(sizeof(char*)));
    TAE(*string=malloc(size+1));
    for (i=0;i<size;++i) {
        (*string)[i]=(i%255)+1;
    }
    (*string)[i]=0;
    return string;
}

static void *create_any(tsf_buffer_t *sub) {
    tsf_buffer_t **result;
    TAE(result=malloc(sizeof(tsf_buffer_t*)));
    *result=sub;
    return result;
}

static void *create_array(tsf_type_t *ele_type,
                          uint32_t total_elements,
                          uint32_t num_elements,
                          void **ele) {
    uint32_t i,j;
    switch (tsf_type_get_kind_code(ele_type)) {
    case TSF_TK_BIT: {
        tsf_native_bitvector_t *result;
        TAE(result=malloc(sizeof(tsf_native_bitvector_t)));
        TAE(result->bits=malloc((total_elements+7)>>3));
        result->num_bits=total_elements;
        for (i=0,j=0;j<total_elements;++j) {
            if (*((tsf_bool_t*)ele[i])) {
                result->bits[j>>3]|=(1<<(j&7));
            } else {
                result->bits[j>>3]&=~(1<<(j&7));
            }
            i=((i+1)%num_elements);
        }
        return result;
    }
    case TSF_TK_VOID: {
        tsf_native_void_array_t *result;
        TAE(result=malloc(sizeof(tsf_native_void_array_t)));
        result->len=total_elements;
        return result;
    }
    default: {
        tsf_native_array_t *result;
        uint8_t *ptr;
        uint32_t ele_size;
        TAT(ele_type=tsf_type_dup(ele_type));
        TAT(tsf_native_type_produce_some_mapping(&ele_type,tsf_true));
        ele_size=tsf_native_type_get_size(ele_type);
        TAE(result=malloc(sizeof(tsf_native_array_t)));
        TAE(result->data=malloc(ele_size*total_elements));
        result->len=total_elements;
        ptr=result->data;
        for (i=0,j=0;j<total_elements;++j) {
            memcpy(ptr,ele[i],ele_size);
            ptr+=ele_size;
            i=((i+1)%num_elements);
        }
        return result;
    }}
}

static void *create_struct(tsf_type_t *type,
                           void **ele) {
    uint8_t *result;
    uint32_t i;
    TAT(type=tsf_type_dup(type));
    TAT(tsf_native_type_produce_some_mapping(&type,tsf_true));
    TAE(result=malloc(tsf_native_type_get_size(type)));
    for (i=0;
         i<tsf_struct_type_get_num_elements(type);
         ++i) {
        tsf_named_type_t *nt=
            tsf_struct_type_get_element(type,i);
        memcpy(result + tsf_native_named_type_get_offset(nt),
               ele[i],
               tsf_native_type_get_size(tsf_named_type_get_type(nt)));
    }
    return result;
}

static void *create_choice(tsf_type_t *type,
                           uint32_t value,
                           void *data) {
    uint8_t *result;
    TAT(type=tsf_type_dup(type));
    TAT(tsf_native_type_produce_some_mapping(&type,tsf_true));
    TAE(result=malloc(tsf_native_type_get_size(type)));
    *((uint32_t*)(result+
                  tsf_native_choice_type_get_value_offset(type)))=
        value;
    if (value!=UINT32_MAX) {
        size_t size=tsf_native_type_get_size
            (tsf_choice_type_get_element_type(type,value));
        memcpy(result+
               tsf_native_choice_type_get_data_offset(type),
               data,
               size);
    }
    return result;
}

/* FIXME: the current approach is that the code in here uses
   tsf_native_type_produce_some_mapping().  alternatively,
   pipe_test_post.h could use the regular native mapping functions.
   this would be harmless in the reflect case, but would make
   the generation case more realistic. */


