/*
 * Copyright (C) 2014 Filip Pizlo. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY FILIP PIZLO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL FILIP PIZLO OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "tsf_internal.h"
#include "tsf_test.h"
#include "tsf_build_defines.h"
#include "gpc.h"

/* the filename is relative to the source directory. the type becomes owned by this function. */
static void *read(const char *filename, tsf_type_t *type) {
    char *real_filename;
    tsf_stream_file_input_t *input;
    tsf_parser_t *parser;
    void *result;
    tsf_buffer_t *buf;
    
    real_filename = tsf_asprintf("%s/%s", SRCDIR, filename);
    
    TAT(input = tsf_stream_file_input_open(real_filename, NULL, NULL));
    free(real_filename);
    TAT(buf = tsf_stream_file_input_read(input));
    tsf_stream_file_input_close(input);
    
    TAT(parser = tsf_parser_create(type));
    TAT(result = tsf_parser_parse(parser, buf));
    tsf_buffer_destroy(buf);
    tsf_parser_destroy(parser);
    
    tsf_type_destroy(type);
    
    return result;
}

typedef struct {
    tsf_choice_t value;
} legacy_choice_t;

typedef struct {
    legacy_choice_t *data;
    uint32_t len;
} legacy_choice_array_t;

static void test_legacy_choice(void) {
    static const unsigned num_choices = 40000;
    
    tsf_type_t *type;
    tsf_type_t *array_type;
    unsigned choice_index;
    legacy_choice_array_t *result;

    TAT(type = tsf_type_create(TSF_TK_CHOICE));
    for (choice_index = 0; choice_index < num_choices; ++choice_index) {
        char *name = tsf_asprintf("%u", choice_index);
        TAT(tsf_choice_type_append(&type, name, tsf_type_create(TSF_TK_VOID)));
        free(name);
    }
    
    TAT(tsf_native_choice_type_map(
            &type, tsf_true,
            tsf_offsetof(legacy_choice_t, value), 0, sizeof(legacy_choice_t)));
    
    TAT(array_type = tsf_type_create_array(type));
    
    result = read("tests/model/plain/model_95b45505f0b9fc0bcc0abe0d0365a3f5b3295a37.tsf", array_type);
    
    TA(result->len == 9);
    TA(result->data[0].value == UINT32_MAX);
    TA(result->data[1].value == 0);
    TA(result->data[2].value == 1);
    TA(result->data[3].value == 127);
    TA(result->data[4].value == 128);
    TA(result->data[5].value == 32383);
    TA(result->data[6].value == 32384);
    TA(result->data[7].value == 32384);
    TA(result->data[8].value == 39999);
    
    tsf_region_free(result);
}

static void do_all_tests(void) {
    DO_TEST(test_legacy_choice());
}

int main(int c, char **v) {
    TEST_INIT(c, v);
    
    do_all_tests();

    TEST_DONE();
    return 0;
}

