/*
 * Copyright (C) 2015 Filip Pizlo. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY FILIP PIZLO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL FILIP PIZLO OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "tsf.h"
#include "tsf_test.h"

static tsf_bool_t use_copier;

static void flip(tsf_type_t *type, void *input, void *output) {
    if (use_copier) {
        tsf_copier_t *copier;
        TAT(copier = tsf_copier_create(type, type));
        TAT(tsf_copier_copy(copier, output, input, NULL));
        tsf_copier_destroy(copier);
    } else {
        tsf_buffer_t *buf;
        tsf_genrtr_t *gen;
        tsf_parser_t *parser;
        
        TAT(gen = tsf_generator_create(type));
        TAT(parser = tsf_parser_create(type));
        
        TAT(buf = tsf_generator_generate(gen, input));
        TAT(tsf_parser_parse_into(parser, buf, output, NULL));
        
        tsf_buffer_destroy(buf);
        tsf_generator_destroy(gen);
        tsf_parser_destroy(parser);
    }
}

static void test_bitvector(void) {
    tsf_type_t *type;
    tsf_native_bitvector_t bitvector;
    tsf_native_bitvector_t bitvector2;
    size_t index;
    
    TAT(type = tsf_type_create_array(tsf_type_create(TSF_TK_BIT)));
    
    bitvector.num_bits = 100;
    TAE(bitvector.bits = malloc(tsf_sizeof_bitvector(bitvector.num_bits)));
    for (index = 0; index < bitvector.num_bits; ++index)
        tsf_bitvector_set(&bitvector, index, (index % 3) != 0);
    
    flip(type, &bitvector, &bitvector2);
    free(bitvector.bits);
    
    TA(bitvector2.num_bits == 100);
    for (index = 0; index < bitvector2.num_bits; ++index)
        TA(tsf_bitvector_get(&bitvector2, index) == ((index % 3) != 0));
    
    tsf_type_destroy(type);
    free(bitvector2.bits);
}

static void test_byte_array(void) {
    tsf_type_t *type;
    tsf_native_int8_array_t byte_array;
    tsf_native_int8_array_t byte_array2;
    size_t index;
    
    TAT(type = tsf_type_create_array(tsf_type_create(TSF_TK_INT8)));
    
    byte_array.len = 100;
    TAE(byte_array.data = malloc(byte_array.len));
    for (index = 0; index < byte_array.len; ++index)
        byte_array.data[index] = index;

    flip(type, &byte_array, &byte_array2);
    free(byte_array.data);
    
    TA(byte_array2.len == 100);
    for (index = 0; index < byte_array2.len; ++index)
        TA(byte_array2.data[index] == index);
    
    tsf_type_destroy(type);
    free(byte_array2.data);
}

static void test_string(const char* the_string) {
    tsf_type_t *type;
    char *string;
    char *string2;
    
    TAT(type = tsf_type_create(TSF_TK_STRING));

    string = (char*)the_string;

    flip(type, &string, &string2);
    
    TA(!strcmp(string2, the_string));
    
    tsf_type_destroy(type);
    free(string2);
}

typedef struct {
    tsf_choice_t value;
    int32_t *number;
} Choice;

static void test_out_of_place_choice(void) {
    tsf_type_t *type;
    Choice choice;
    Choice choice2;
    int32_t number;
    
    TAT(type = tsf_native_choice_type_create(tsf_false,
                                             tsf_offsetof(Choice, value),
                                             tsf_offsetof(Choice, number),
                                             sizeof(Choice)));
    TAT(tsf_choice_type_append(&type, "thingy", tsf_type_create(TSF_TK_INT32)));
    
    choice.value = 0;
    choice.number = &number;
    number = 47;

    flip(type, &choice, &choice2);
    
    TA(choice2.value == 0);
    TA(choice2.number != &number);
    TA(*choice2.number == 47);
    
    free(choice2.number);
    
    tsf_type_destroy(type);
}

typedef struct {
    tsf_integer_t *data;
    uint32_t len;
} integer__array_t;

static void test_integer_array(void) {
    tsf_type_t *type;
    integer__array_t array;
    integer__array_t array2;
    unsigned i;
    
    TAT(type = tsf_type_create_array(tsf_type_create(TSF_TK_INTEGER)));
    
    array.len = 100;
    TAE(array.data = malloc(sizeof(tsf_integer_t) * array.len));
    for (i = array.len; i--;)
        array.data[i] = i * 1000;
    
    flip(type, &array, &array2);
    
    TA(array2.len == array.len);
    for (i = array.len; i--;)
        TA(array.data[i] == array2.data[i]);
    
    free(array.data);
    free(array2.data);
    tsf_type_destroy(type);
}

typedef struct {
    tsf_native_int32_array_t *array;
    uint32_t len;
} int32_array__array_t;

static void destroy_int32_array_array(int32_array__array_t *array) {
    uint32_t i;
    for (i = array->len; i--;)
        free(array->array[i].data);
    free(array->array);
}

static void test_int32_array_array(void) {
    tsf_type_t *type;
    int32_array__array_t array;
    int32_array__array_t array2;
    unsigned i, j;
    
    TAT(type = tsf_type_create_array(tsf_type_create_array(tsf_type_create(TSF_TK_INT32))));
    
    array.len = 100;
    TAE(array.array = malloc(sizeof(int32_array__array_t) * array.len));
    for (i = array.len; i--;) {
        tsf_native_int32_array_t *sub_array = array.array + i;
        sub_array->len = i;
        TAE(sub_array->data = malloc(sizeof(int32_t) * sub_array->len));
        for (j = sub_array->len; j--;)
            sub_array->data[j] = i + j;
    }
    
    flip(type, &array, &array2);
    
    TA(array.len == array2.len);
    for (i = array.len; i--;) {
        tsf_native_int32_array_t *sub_array = array.array + i;
        tsf_native_int32_array_t *sub_array2 = array2.array + i;
        TA(sub_array->len == sub_array2->len);
        for (j = sub_array->len; j--;)
            TA(sub_array->data[j] == sub_array2->data[j]);
    }
    
    destroy_int32_array_array(&array);
    destroy_int32_array_array(&array2);
    tsf_type_destroy(type);
}

static void do_all_tests(void) {
    DO_TEST(test_bitvector());
    DO_TEST(test_byte_array());
    DO_TEST(test_string(""));
    DO_TEST(test_string("hello world"));
    DO_TEST(test_string("hello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello worldhello world")); /* test a long string to force realloc */
    DO_TEST(test_out_of_place_choice());
    DO_TEST(test_integer_array());
    DO_TEST(test_int32_array_array());
}

int main(int c, char **v) {
    TEST_INIT(c, v);
    
    use_copier = tsf_false;
    TEST_PUSH("generator_to_parser");
    do_all_tests();
    TEST_POP();
    
    use_copier = tsf_true;
    TEST_PUSH("copier");
    do_all_tests();
    TEST_POP();
    
    TEST_DONE();
    
    return 0;
}

