#include "proto_tuple.pb.h"
#include <iostream>
#include <fcntl.h>
#include <fstream>
#include <string>
#include <sys/time.h>
#include <tsf/tsf.h>

using namespace std;

static double milliTime() {
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return tv.tv_sec * 1000. + tv.tv_usec / 1000.;
}

#define TIMEIT(statement) do {                                          \
        double __ti_before = milliTime();                               \
        statement;                                                      \
        double __ti_after = milliTime();                                \
        printf("%s: %.3lf ms\n", #statement, __ti_after - __ti_before); \
    } while (0)

/* If the expression evaluates false then it assumes an error and exits. */
#define CT(exp) do {                                                    \
        if (!(exp)) {                                                   \
            fprintf(stderr, "%s:%d: %s: %s\n",                          \
                    __FILE__, __LINE__, #exp, tsf_get_error());         \
            exit(1);                                                    \
        }                                                               \
    } while (0)
#define CS(exp) do {                                                    \
        if (!(exp)) {                                                   \
            int myErrno = errno;                                        \
            fprintf(stderr, "%s:%d: %s: %s\n",                          \
                    __FILE__, __LINE__, #exp, strerror(myErrno));       \
            exit(1);                                                    \
        }                                                               \
    } while (0)

static const char *filename = "proto_tuple_speed_test_file.proto";

static void writeTest(unsigned numTuples) {
    int fd;
    tsf_buf_wtr_t *writer;
    char buffer[1000];
    
    CS((fd = open(filename, O_CREAT | O_TRUNC | O_WRONLY, 0644)) >= 0);
    
    CT(writer = tsf_buf_writer_create(fd, TSF_DEF_BUF_SIZE, tsf_true));

    for (unsigned tupleIndex = 0; tupleIndex < numTuples; ++tupleIndex) {
        Tuple tuple;
        tuple.set_foo(tupleIndex);
        tuple.set_bar(tupleIndex + 1);
        tuple.set_baz(tupleIndex + 2);
        tuple.set_thingy(tupleIndex + 3);
        
        uint32_t length = tuple.ByteSize();
        if (length > sizeof(buffer))
            abort();
        tuple.SerializeToArray(buffer, length);
        
        CT(tsf_buf_writer_write(writer, &length, sizeof(length)));
        CT(tsf_buf_writer_write(writer, (void*)buffer, length));
    }
    
    CT(tsf_buf_writer_flush(writer));
    tsf_buf_writer_destroy(writer);
}

static void readTest(unsigned numTuples) {
    int fd;
    tsf_buf_rdr_t *reader;
    char buffer[1000];
    
    CS((fd = open(filename, O_RDONLY)) >= 0);
    CT(reader = tsf_buf_reader_create(fd, TSF_DEF_BUF_SIZE, tsf_true));
    
    Tuple tuple;
    unsigned count = 0;
    uint64_t total = 0;
    for (;;) {
        uint32_t length;
        if (!tsf_buf_reader_read(reader, &length, sizeof(length))) {
            CT(tsf_get_error_code() == TSF_E_UNEXPECTED_EOF);
            break;
        }

        if (length > sizeof(buffer))
            abort();
        CT(tsf_buf_reader_read(reader, buffer, length));
        
        tuple.ParseFromArray(buffer, length);
        
        total += (uint64_t)tuple.foo();
        total += (uint64_t)tuple.bar();
        total += (uint64_t)tuple.baz();
        total += (uint64_t)tuple.thingy();
        
        count++;
    }

    if (count != numTuples) {
        fprintf(stderr, "Wrong number of tuples read: %u, expected %u.\n", numTuples, count);
        exit(1);
    }

    uint64_t expectedTotal = ((((uint64_t)count - 1) * count) / 2) * 4 + (1 + 2 + 3) * count;
    if (total != expectedTotal) {
        fprintf(stderr, "Wrong total: %llu, expected %llu.\n", total, expectedTotal);
        exit(1);
    }
    
    tsf_buf_reader_destroy(reader);
}

int main(int c, char **v) {
    GOOGLE_PROTOBUF_VERIFY_VERSION;
    
    TIMEIT(writeTest(1000000));
    TIMEIT(readTest(1000000));
    
    return 0;
}

