/*
 * Copyright (C) 2003, 2004, 2005, 2014 Filip Pizlo. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY FILIP PIZLO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL FILIP PIZLO OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#ifndef FP_TEST_H
#define FP_TEST_H

#include "tsf.h"
#include "tsf_format.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

void tsf_test_push(const char *msg,...);
void tsf_test_pop(void);

char *tsf_test_current(void);   /* returns a new string each time */

void tsf_do_test_first_hook(void);  /* first execution of the test */
tsf_bool_t tsf_do_test_pre_hook(void);  /* returns true if the test should be run */
tsf_bool_t tsf_do_test_loop_hook(void); /* returns true if the test should be run again */

void tsf_test_init(int c,char **v,tsf_bool_t do_forking);
void tsf_test_report(void);
void tsf_test_done(void);

tsf_bool_t tsf_should_run_test(void);

typedef enum {
    /* the test succeeded */
    tsf_test_success,
    
    /* the test failed, either because of TEST_FAIL() or a failed
     * assertion */
    tsf_test_failure,
    
    /* was expecting to see an error propagated, but it wasn't */
    tsf_test_error_dropped,
    
    /* a system error occurred */
    tsf_test_system_error,
    
    /* a TSF error occurred */
    tsf_test_tsf_error,
    
    /* the child exited prematurely */
    tsf_test_exit,
    
    /* the child crashed */
    tsf_test_crash,
} tsf_test_status_t;

typedef enum {
    /* no error */
    tsf_error_none,
    
    /* there is a system error (errno) */
    tsf_error_system,
    
    /* there is a TSF error */
    tsf_error_tsf
    
} tsf_error_status_t;

void tsf_test_complete(tsf_test_status_t status,
                       tsf_error_status_t error_status,
                       const char *format,
                       ...);

/* breakpoint function.  if you tell the test suite to break on a certain
 * test case, it will do so using the breakpoint function.  of course, the
 * breakpoint function doesn't actually do anything.  so, it is your
 * responsibility to set a breakpoint on the breakpoint function using
 * your debugger. */
void tsf_test_break() TSF_ATTRIBUTE_NOINLINE;

#if 0
} /* unconfuse emacs */
#endif

/* In general, the APIs that you want to use are the ones in ALL CAPS.
 * The other ones may be of use if you're instrumenting the test suite or
 * something. */

#define TEST_PUSH tsf_test_push
#define TEST_POP tsf_test_pop

#define TEST_INIT(c,v) do {\
    tsf_test_init(c,v,tsf_true);\
    fprintf(stderr,"%s: ",v[0]);\
    TEST_PUSH(v[0]);\
} while (0)

#define TEST_INIT_NOFORK(c,v) do {\
    tsf_test_init(c,v,tsf_false);\
    fprintf(stderr,"%s: ",v[0]);\
    TEST_PUSH(v[0]);\
} while (0)

#define TEST_DONE_CUSTOM(msg) \
    (fprintf(stderr,"%s\n",msg))

#define TEST_DONE() do {\
    tsf_test_report();\
    tsf_test_done();\
} while (0)

#define TEST_FAIL(reason) do {\
    tsf_test_complete(tsf_test_failure,\
                      tsf_error_none,\
                      "Failure at %s:%d:\n\t%s\n",\
                      __FILE__,__LINE__,reason);\
    abort(); /* never actually called, but I use this to hint any */ \
             /* dataflow analysis that the C compiler is doing */ \
} while (0)

#define TF(reason) TEST_FAIL(reason)

#define TEST_ASSERT(exp) do {\
    if (!(exp)) {\
        tsf_test_complete(tsf_test_failure,\
                          tsf_error_none,\
                          "TEST_ASSERT(%s) failed at %s:%d.\n",\
                          #exp,__FILE__,__LINE__);\
        abort(); /* never actually called, but I use this to hint any */ \
                 /* dataflow analysis that the C compiler is doing */ \
    }\
} while (0)

#define TA(exp) TEST_ASSERT(exp)

#define TEST_ASSERT_STR_EQUAL(_expected,_gotten) do {		\
	const char *__tase_expected = (_expected);              \
	const char *__tase_gotten = (_gotten);                  \
	if ((__tase_expected==NULL && __tase_gotten!=NULL) ||	\
	    (__tase_gotten==NULL && __tase_expected!=NULL) ||	\
	    (__tase_expected!=NULL && strcmp(__tase_expected, __tase_gotten))) { \
	    tsf_test_complete(tsf_test_failure,			\
			      tsf_error_none,			\
			      "TEST_ASSERT_STR_EQUAL"		\
			      "(expected: %s = %s, "		\
			      "got: %s = %s) "			\
			      "failed at %s:%d.\n",		\
			      #_expected,__tase_expected,	\
			      #_gotten,__tase_gotten,	        \
			      __FILE__,__LINE__);		\
	    abort();						\
	}							\
    } while (0)

#define TASE(expected,gotten) TEST_ASSERT_STR_EQUAL(expected,gotten)

#define TEST_ASSERT_ERRNO(exp) do {\
    if (!(exp)) {\
        tsf_test_complete(tsf_test_system_error,\
                          tsf_error_system,\
                          "TEST_ASSERT_ERRNO(%s) failed at %s:%d with error:\n"\
                          "\t%s\n",\
                          #exp,__FILE__,__LINE__,strerror(errno));\
        abort(); /* never actually called, but I use this to hint any */ \
                 /* dataflow analysis that the C compiler is doing */ \
    }\
} while (0)

#define TAE(exp) TEST_ASSERT_ERRNO(exp)

#define TEST_EXPECT_ERRNO(exp,expected_errno) do {\
    if (!(exp)) {\
        if (errno!=expected_errno) {\
            tsf_test_complete(tsf_test_failure,\
                              tsf_error_system,\
                              "TEST_EXPECT_ERRNO(%s, %s) failed with wrong "\
                              "error at %s:%d:\n\t%s",\
                              #exp,strerror(expected_errno),\
                              __FILE__,__LINE__,\
                              strerror(errno));\
            abort(); /* never actually called, but I use this to hint any */ \
                     /* dataflow analysis that the C compiler is doing */ \
        }\
    } else {\
        tsf_test_complete(tsf_test_failure,\
                          tsf_error_none,\
                          "TEST_EXPECT_ERRNO(%s, %s) failed to fail at %s:%d.\n",\
                          #exp,strerror(expected_errno),__FILE__,__LINE__);\
        abort(); /* never actually called, but I use this to hint any */ \
                 /* dataflow analysis that the C compiler is doing */ \
    }\
} while (0)

#define TEE(exp,expected_errno) TEST_EXPECT_ERRNO(exp,expected_errno)

#define TEST_ASSERT_TSFERROR(exp) do {\
    if (!(exp)) {\
        tsf_test_complete(tsf_test_tsf_error,\
                          tsf_error_tsf,\
                          "TEST_ASSERT_TSFERROR(%s) failed at %s:%d with "\
                          "error:\n\t%s\n",\
                          #exp,__FILE__,__LINE__,tsf_get_error());\
        abort(); /* never actually called, but I use this to hint any */ \
                 /* dataflow analysis that the C compiler is doing */ \
    }\
} while (0)

#define TAT(exp) TEST_ASSERT_TSFERROR(exp)

#define TEST_EXPECT_TSFERROR(exp,expected_error) do {\
    if (!(exp)) {\
        if (tsf_get_error_code()!=expected_error) {\
            tsf_test_complete(tsf_test_failure,\
                              tsf_error_tsf,\
                              "TEST_EXPECT_TSFERROR(%s, %s) failed with wrong "\
                              "error at %s:%d:\n\t%s",\
                              #exp,\
                              tsf_get_message_for_error_code(expected_error),\
                              __FILE__,__LINE__,\
                              tsf_get_error());\
            abort(); /* never actually called, but I use this to hint any */ \
                     /* dataflow analysis that the C compiler is doing */ \
        }\
    } else {\
        tsf_test_complete(tsf_test_failure,\
                          tsf_error_none,\
                          "TEST_EXPECT_TSFERROR(%s, %s) failed to fail at "\
                          "%s:%d.\n",#exp,\
                          tsf_get_message_for_error_code(expected_error),\
                          __FILE__,__LINE__);\
        abort(); /* never actually called, but I use this to hint any */ \
                 /* dataflow analysis that the C compiler is doing */ \
    }\
} while (0)

#define TET(exp,expected_errno) TEST_EXPECT_TSFERROR(exp,expected_errno)

#define DO_TEST(test) do {                                              \
        TEST_PUSH("%s",#test);                                          \
        if (tsf_should_run_test()) {                                    \
            tsf_do_test_first_hook();                                   \
            do {                                                        \
                if (tsf_do_test_pre_hook()) {                           \
                    test;                                               \
                    tsf_test_complete(tsf_test_success,                 \
                                      tsf_error_none,                   \
                                      "Success\n");                     \
                }                                                       \
            } while (tsf_do_test_loop_hook());                          \
        }                                                               \
        TEST_POP();                                                     \
    } while (0)

#define REPEAT(test,n) do {\
    uint32_t i;\
    for (i=0;i<(n);++i) {\
        test;\
    }\
} while (0)

#endif

