/*
 * Copyright (C) 2004, 2005 Filip Pizlo. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY FILIP PIZLO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL FILIP PIZLO OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "tsf_internal.h"
#include "tsf_test.h"
#include "tsf_format.h"

typedef void (*typed_test_t)(tsf_type_t *type);

/* writes the type out and then reads it with limits set so that if the
 * type was one byte longer, it wouldn't parse. */
static void barely_legal(tsf_type_t *type) {
    uint32_t byte_size;
    tsf_light_buffer_t lb;
    tsf_limits_t *limits;
    tsf_type_t *type2;
    uint8_t *buf;
    
    TAT(type);
    
    byte_size=tsf_type_get_byte_size(type);
    
    TAE(buf=lb.cur=lb.end=malloc(byte_size));
    TAT(tsf_type_write(type,
                       tsf_buffer_only_writer,
                       &(lb.end)));
    TA(lb.cur+byte_size==lb.end);
    
    TAT(limits=tsf_limits_create());
    limits->max_type_size=byte_size;
    
    TAT(type2=tsf_type_read(tsf_buffer_only_reader,
                            &lb,
                            limits));
    TA(tsf_type_compare(type,type2));
    
    free(buf);
    tsf_type_destroy(type);
    tsf_type_destroy(type2);
}

/* writes the type out and then reads it with limits set so that the
 * type should not be parsed */
static void off_by_one(tsf_type_t *type) {
    uint32_t byte_size;
    tsf_light_buffer_t lb;
    tsf_limits_t *limits;
    uint8_t *buf;
    
    TAT(type);
    
    byte_size=tsf_type_get_byte_size(type);
    
    TAE(buf=lb.cur=lb.end=malloc(byte_size));
    TAT(tsf_type_write(type,
                       tsf_buffer_only_writer,
                       &(lb.end)));
    TA(lb.cur+byte_size==lb.end);
    
    TAT(limits=tsf_limits_create());
    limits->max_type_size=byte_size-1;
    
    TET(tsf_type_read(tsf_buffer_only_reader,
                      &lb,
                      limits), TSF_E_PARSE_ERROR);
    
    free(buf);
    tsf_type_destroy(type);
}

/* the different types to play with */
static void byte_limit_type_cases(typed_test_t doit) {
    DO_TEST(doit(tsf_type_create(TSF_TK_VOID)));
    DO_TEST(doit(tsf_type_create(TSF_TK_INT32)));
    DO_TEST(doit(tsf_type_create_struct(NULL)));
    DO_TEST(doit(tsf_type_create_struct("x",
                                        tsf_type_create(TSF_TK_FLOAT),
                                        NULL)));
    DO_TEST(doit(tsf_type_create_struct("x",
                                        tsf_type_create(TSF_TK_FLOAT),
                                        "y",
                                        tsf_type_create(TSF_TK_FLOAT),
                                        NULL)));
    DO_TEST(doit(tsf_type_create_choice(NULL)));
    DO_TEST(doit(tsf_type_create_choice("x",
                                        tsf_type_create(TSF_TK_FLOAT),
                                        NULL)));
    DO_TEST(doit(tsf_type_create_choice("x",
                                        tsf_type_create(TSF_TK_FLOAT),
                                        "y",
                                        tsf_type_create(TSF_TK_FLOAT),
                                        NULL)));
    DO_TEST(doit(tsf_type_create_array(tsf_type_create(TSF_TK_VOID))));
    DO_TEST(doit(tsf_type_create_array(tsf_type_create(TSF_TK_INT32))));
}

/* verify that we can disallow a particular kind type */
static void disallow_type(tsf_type_t *type) {
    uint32_t byte_size;
    tsf_light_buffer_t lb;
    tsf_limits_t *limits;
    uint8_t *buf;
    
    TAT(type);
    
    byte_size=tsf_type_get_byte_size(type);
    
    TAE(buf=lb.cur=lb.end=malloc(byte_size));
    TAT(tsf_type_write(type,
                       tsf_buffer_only_writer,
                       &(lb.end)));
    TA(lb.cur+byte_size==lb.end);
    
    TAT(limits=tsf_limits_create());
    tsf_limits_reset_allow_type_kind(limits,
                                     tsf_type_get_kind_code(type));
    
    TET(tsf_type_read(tsf_buffer_only_reader,
                      &lb,
                      limits), TSF_E_PARSE_ERROR);
    
    free(buf);
    tsf_type_destroy(type);
}

/* type cases for each kind type */
static void type_kind_cases() {
    DO_TEST(disallow_type(tsf_type_create(TSF_TK_VOID)));
    DO_TEST(disallow_type(tsf_type_create(TSF_TK_INT8)));
    DO_TEST(disallow_type(tsf_type_create(TSF_TK_UINT8)));
    DO_TEST(disallow_type(tsf_type_create(TSF_TK_INT16)));
    DO_TEST(disallow_type(tsf_type_create(TSF_TK_UINT16)));
    DO_TEST(disallow_type(tsf_type_create(TSF_TK_INT32)));
    DO_TEST(disallow_type(tsf_type_create(TSF_TK_UINT32)));
    DO_TEST(disallow_type(tsf_type_create(TSF_TK_INT64)));
    DO_TEST(disallow_type(tsf_type_create(TSF_TK_UINT64)));
    DO_TEST(disallow_type(tsf_type_create(TSF_TK_FLOAT)));
    DO_TEST(disallow_type(tsf_type_create(TSF_TK_DOUBLE)));
    DO_TEST(disallow_type(tsf_type_create(TSF_TK_BIT)));
    DO_TEST(disallow_type(tsf_type_create(TSF_TK_STRING)));
    DO_TEST(disallow_type(
        tsf_type_create_array(tsf_type_create(TSF_TK_VOID))));
    DO_TEST(disallow_type(tsf_type_create(TSF_TK_STRUCT)));
    DO_TEST(disallow_type(tsf_type_create(TSF_TK_CHOICE)));
    DO_TEST(disallow_type(tsf_type_create(TSF_TK_ANY)));
}

int main(int c,char **v) {
    TEST_INIT(c,v);
    
    TEST_PUSH("byte_limit");
    TEST_PUSH("doit=barely_legal");
    byte_limit_type_cases(barely_legal);
    TEST_POP();
    
    TEST_PUSH("doit=off_by_one");
    byte_limit_type_cases(off_by_one);
    TEST_POP();
    TEST_POP();
    
    TEST_PUSH("type_kind_limit");
    type_kind_cases();
    TEST_POP();
    
    TEST_DONE();
    
    return 0;
}




